@@ -54,6 +54,9 @@ static void gnutls_debug_log_all(int level, const char *str)
 {
     apr_file_printf(debug_log_fp, "<%d> %s\n", level, str);
 }
+#define _gnutls_log apr_file_printf
+#else
+# define _gnutls_log(...) 
 #endif
 
 int
@@ -62,6 +65,18 @@ mgs_hook_pre_config(apr_pool_t * pconf,
 {
 int ret;
 
+#if MOD_GNUTLS_DEBUG
+    apr_file_open(&debug_log_fp, "/tmp/gnutls_debug",
+		  APR_APPEND | APR_WRITE | APR_CREATE, APR_OS_DEFAULT,
+		  pconf);
+
+    _gnutls_log(debug_log_fp, "%s: %d\n", __func__, __LINE__);
+
+    gnutls_global_set_log_level(9);
+    gnutls_global_set_log_function(gnutls_debug_log_all);
+    _gnutls_log(debug_log_fp, "gnutls: %s\n", gnutls_check_version(NULL));
+#endif
+
 #if APR_HAS_THREADS
     ap_mpm_query(AP_MPMQ_IS_THREADED, &mpm_is_threaded);
     if (mpm_is_threaded) {
@@ -72,29 +87,20 @@ int ret;
 #endif
 
     if (gnutls_check_version(LIBGNUTLS_VERSION)==NULL) {
-        fprintf(stderr, "gnutls_check_version() failed. Required: gnutls-%s Found: gnutls-%s\n", 
+        _gnutls_log(debug_log_fp, "gnutls_check_version() failed. Required: gnutls-%s Found: gnutls-%s\n", 
           LIBGNUTLS_VERSION, gnutls_check_version(NULL));
         return -3;
     }
 
     ret = gnutls_global_init();
     if (ret < 0) {
-        fprintf(stderr, "gnutls_global_init: %s\n", gnutls_strerror(ret));
+        _gnutls_log(debug_log_fp, "gnutls_global_init: %s\n", gnutls_strerror(ret));
         return -3;
     }
 
     apr_pool_cleanup_register(pconf, NULL, mgs_cleanup_pre_config,
 			      apr_pool_cleanup_null);
 
-#if MOD_GNUTLS_DEBUG
-    apr_file_open(&debug_log_fp, "/tmp/gnutls_debug",
-		  APR_APPEND | APR_WRITE | APR_CREATE, APR_OS_DEFAULT,
-		  pconf);
-
-    gnutls_global_set_log_level(9);
-    gnutls_global_set_log_function(gnutls_debug_log_all);
-    apr_file_printf(debug_log_fp, "gnutls: %s\n", gnutls_check_version(NULL));
-#endif
 
     return OK;
 }
@@ -106,6 +112,8 @@ static int mgs_select_virtual_server_cb(gnutls_session_t session)
     int ret;
     int cprio[2];
 
+    _gnutls_log(debug_log_fp, "%s: %d\n", __func__, __LINE__);
+
     ctxt = gnutls_transport_get_ptr(session);
 
     /* find the virtual server */
@@ -115,7 +123,7 @@ static int mgs_select_virtual_server_cb(gnutls_session_t session)
 	ctxt->sc = tsc;
 
     gnutls_certificate_server_set_request(session,
-					  ctxt->sc->client_verify_mode);
+					      ctxt->sc->client_verify_mode);
 
     /* set the new server credentials 
      */
@@ -162,8 +170,12 @@ static int cert_retrieve_fn(gnutls_session_t session, gnutls_retr_st * ret)
 {
     mgs_handle_t *ctxt;
 
+    _gnutls_log(debug_log_fp, "%s: %d\n", __func__, __LINE__);
     ctxt = gnutls_transport_get_ptr(session);
 
+    if (ctxt == NULL)
+        return GNUTLS_E_INTERNAL_ERROR;
+
     if (gnutls_certificate_type_get( session) == GNUTLS_CRT_X509) {
         ret->type = GNUTLS_CRT_X509;
         ret->ncerts = ctxt->sc->certs_x509_num;
@@ -210,6 +222,7 @@ static int read_crt_cn(server_rec * s, apr_pool_t * p,
     size_t data_len;
 
 
+    _gnutls_log(debug_log_fp,   "%s: %d\n", __func__, __LINE__);
     *cert_cn = NULL;
 
     data_len = 0;
@@ -261,6 +274,7 @@ static int read_pgpcrt_cn(server_rec * s, apr_pool_t * p,
     size_t data_len;
 
 
+    _gnutls_log(debug_log_fp, "%s: %d\n", __func__, __LINE__);
     *cert_cn = NULL;
 
     data_len = 0;
@@ -293,6 +307,7 @@ mgs_hook_post_config(apr_pool_t * p, apr_pool_t * plog,
     int first_run = 0;
     const char *userdata_key = "mgs_init";
 
+    _gnutls_log(debug_log_fp, "%s: %d\n", __func__, __LINE__);
     apr_pool_userdata_get(&data, userdata_key, base_server->process->pool);
     if (data == NULL) {
 	first_run = 1;
@@ -385,7 +400,7 @@ mgs_hook_post_config(apr_pool_t * p, apr_pool_t * plog,
 							    srp_tpasswd_conf_file);
 
 		if (rv < 0 && sc->enabled == GNUTLS_ENABLED_TRUE) {
-		    ap_log_error(APLOG_MARK, APLOG_EMERG, 0, s,
+		    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
 				 "[GnuTLS] - Host '%s:%d' is missing a "
 				 "SRP password or conf File!",
 				 s->server_hostname, s->port);
@@ -394,18 +409,20 @@ mgs_hook_post_config(apr_pool_t * p, apr_pool_t * plog,
 	    }
 #endif
 
-	    if (sc->certs_x509[0] == NULL
-		&& sc->enabled == GNUTLS_ENABLED_TRUE) {
-		ap_log_error(APLOG_MARK, APLOG_EMERG, 0, s,
+	    if (sc->certs_x509[0] == NULL &&
+	        sc->cert_pgp == NULL &&
+		sc->enabled == GNUTLS_ENABLED_TRUE) {
+		ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
 			     "[GnuTLS] - Host '%s:%d' is missing a "
 			     "Certificate File!", s->server_hostname,
 			     s->port);
 		exit(-1);
 	    }
 
-	    if (sc->privkey_x509 == NULL
-		&& sc->enabled == GNUTLS_ENABLED_TRUE) {
-		ap_log_error(APLOG_MARK, APLOG_EMERG, 0, s,
+	    if (sc->enabled == GNUTLS_ENABLED_TRUE && 
+	      ((sc->certs_x509[0] != NULL && sc->privkey_x509 == NULL) ||
+	      (sc->cert_pgp != NULL && sc->privkey_pgp == NULL))) {
+		ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
 			     "[GnuTLS] - Host '%s:%d' is missing a "
 			     "Private Key File!",
 			     s->server_hostname, s->port);
@@ -418,7 +435,7 @@ mgs_hook_post_config(apr_pool_t * p, apr_pool_t * plog,
     		    rv = read_pgpcrt_cn(s, p, sc->cert_pgp, &sc->cert_cn);
 
 		if (rv < 0) {
-		    ap_log_error(APLOG_MARK, APLOG_EMERG, 0, s,
+		    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
 				 "[GnuTLS] - Cannot find a certificate for host '%s:%d'!",
 				 s->server_hostname, s->port);
 		    sc->cert_cn = NULL;
@@ -439,6 +456,7 @@ void mgs_hook_child_init(apr_pool_t * p, server_rec * s)
     mgs_srvconf_rec *sc = ap_get_module_config(s->module_config,
 					       &gnutls_module);
 
+    _gnutls_log(debug_log_fp,   "%s: %d\n", __func__, __LINE__);
     if (sc->cache_type != mgs_cache_none) {
 	rv = mgs_cache_child_init(p, s, sc);
 	if (rv != APR_SUCCESS) {
@@ -457,6 +475,7 @@ const char *mgs_hook_http_scheme(const request_rec * r)
 	(mgs_srvconf_rec *) ap_get_module_config(r->server->module_config,
 						 &gnutls_module);
 
+    _gnutls_log(debug_log_fp,   "%s: %d\n", __func__, __LINE__);
     if (sc->enabled == GNUTLS_ENABLED_FALSE) {
 	return NULL;
     }
@@ -470,6 +489,7 @@ apr_port_t mgs_hook_default_port(const request_rec * r)
 	(mgs_srvconf_rec *) ap_get_module_config(r->server->module_config,
 						 &gnutls_module);
 
+    _gnutls_log(debug_log_fp,   "%s: %d\n", __func__, __LINE__);
     if (sc->enabled == GNUTLS_ENABLED_FALSE) {
 	return 0;
     }
@@ -491,6 +511,7 @@ static int vhost_cb(void *baton, conn_rec * conn, server_rec * s)
     mgs_srvconf_rec *tsc;
     vhost_cb_rec *x = baton;
 
+    _gnutls_log(debug_log_fp,   "%s: %d\n", __func__, __LINE__);
     tsc = (mgs_srvconf_rec *) ap_get_module_config(s->module_config,
 						   &gnutls_module);
 
@@ -543,6 +564,7 @@ mgs_srvconf_rec *mgs_find_sni_server(gnutls_session_t session)
     mgs_srvconf_rec *tsc;
 #endif
 
+    _gnutls_log(debug_log_fp,   "%s: %d\n", __func__, __LINE__);
     ctxt = gnutls_transport_get_ptr(session);
 
     rv = gnutls_server_name_get(ctxt->session, sni_name,
@@ -620,6 +642,7 @@ static mgs_handle_t *create_gnutls_handle(apr_pool_t * pool, conn_rec * c)
 						 module_config,
 						 &gnutls_module);
 
+    _gnutls_log(debug_log_fp,   "%s: %d\n", __func__, __LINE__);
     ctxt = apr_pcalloc(pool, sizeof(*ctxt));
     ctxt->c = c;
     ctxt->sc = sc;
@@ -658,10 +681,15 @@ int mgs_hook_pre_connection(conn_rec * c, void *csd)
 						 module_config,
 						 &gnutls_module);
 
+    _gnutls_log(debug_log_fp,   "%s: %d\n", __func__, __LINE__);
     if (!(sc && (sc->enabled == GNUTLS_ENABLED_TRUE))) {
 	return DECLINED;
     }
 
+    if(c->remote_addr->hostname)
+      /* Connection initiated by Apache (mod_proxy) => ignore */
+      return OK;
+
     ctxt = create_gnutls_handle(c->pool, c);
 
     ap_set_module_config(c->conn_config, &gnutls_module, ctxt);
@@ -687,6 +715,7 @@ int mgs_hook_fixups(request_rec * r)
     mgs_handle_t *ctxt;
     int rv = OK;
 
+    _gnutls_log(debug_log_fp,   "%s: %d\n", __func__, __LINE__);
     apr_table_t *env = r->subprocess_env;
 
     ctxt =
@@ -721,8 +750,8 @@ int mgs_hook_fixups(request_rec * r)
 					       (ctxt->session)));
 
 #ifdef ENABLE_SRP
-    apr_table_setn(env, "SSL_SRP_USER",
-		   gnutls_srp_server_get_username(ctxt->session));
+    tmp = gnutls_srp_server_get_username(ctxt->session);
+    apr_table_setn(env, "SSL_SRP_USER", (tmp!=NULL)?tmp:"");
 #endif
 
     if (apr_table_get(env, "SSL_CLIENT_VERIFY") == NULL)
@@ -761,6 +790,7 @@ int mgs_hook_authz(request_rec * r)
     mgs_dirconf_rec *dc = ap_get_module_config(r->per_dir_config,
 					       &gnutls_module);
 
+    _gnutls_log(debug_log_fp,   "%s: %d\n", __func__, __LINE__);
     ctxt =
 	ap_get_module_config(r->connection->conn_config, &gnutls_module);
 
@@ -778,6 +808,12 @@ int mgs_hook_authz(request_rec * r)
 			  ctxt->sc->client_verify_mode,
 			  dc->client_verify_mode);
 
+            /* If we already have a client certificate, there's no point in
+             * re-handshaking... */
+            rv = mgs_cert_verify(r, ctxt);
+            if (rv != DECLINED && rv != HTTP_FORBIDDEN)
+                return rv;
+
 	    gnutls_certificate_server_set_request(ctxt->session,
 						  dc->client_verify_mode);
 
@@ -789,11 +825,13 @@ int mgs_hook_authz(request_rec * r)
 	    ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r,
 			  "GnuTLS: Peer is set to IGNORE");
 #endif
-	} else {
-	    rv = mgs_cert_verify(r, ctxt);
-	    if (rv != DECLINED) {
-		return rv;
-	    }
+	    return DECLINED;
+	}
+	rv = mgs_cert_verify(r, ctxt);
+	if (rv != DECLINED &&
+	    (rv != HTTP_FORBIDDEN ||
+	     dc->client_verify_mode == GNUTLS_CERT_REQUIRE)) {
+	    return rv;
 	}
     }
 
@@ -822,6 +860,7 @@ mgs_add_common_cert_vars(request_rec * r, gnutls_x509_crt_t cert, int side,
 
     apr_table_t *env = r->subprocess_env;
 
+    _gnutls_log(debug_log_fp,   "%s: %d\n", __func__, __LINE__);
     if (export_certificates_enabled != 0) {
 	char cert_buf[10 * 1024];
 	len = sizeof(cert_buf);
@@ -928,6 +967,7 @@ mgs_add_common_pgpcert_vars(request_rec * r, gnutls_openpgp_crt_t cert, int side
     size_t len;
     int ret;
 
+    _gnutls_log(debug_log_fp,   "%s: %d\n", __func__, __LINE__);
     apr_table_t *env = r->subprocess_env;
 
     if (export_certificates_enabled != 0) {
@@ -988,12 +1028,14 @@ static int mgs_cert_verify(request_rec * r, mgs_handle_t * ctxt)
     const gnutls_datum_t *cert_list;
     unsigned int cert_list_size, status, expired;
     int rv, ret;
+    unsigned int ch_size = 0;
     union {
-      gnutls_x509_crt_t x509;
+      gnutls_x509_crt_t x509[MAX_CHAIN_SIZE];
       gnutls_openpgp_crt_t pgp;
     } cert;
     apr_time_t activation_time, expiration_time, cur_time;
 
+    _gnutls_log(debug_log_fp, "%s: %d\n", __func__, __LINE__);
     cert_list =
 	gnutls_certificate_get_peers(ctxt->session, &cert_list_size);
 
@@ -1010,19 +1052,40 @@ static int mgs_cert_verify(request_rec * r, mgs_handle_t * ctxt)
 	return HTTP_FORBIDDEN;
     }
 
-    if (cert_list_size > 1) {
-	ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r,
+    if (gnutls_certificate_type_get( ctxt->session) == GNUTLS_CRT_X509) {
+        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+            "GnuTLS: A Chain of %d certificate(s) was provided for validation", cert_list_size);
+
+        for (ch_size =0; ch_size<cert_list_size; ch_size++) {
+            gnutls_x509_crt_init(&cert.x509[ch_size]);
+            rv = gnutls_x509_crt_import(cert.x509[ch_size], &cert_list[ch_size], GNUTLS_X509_FMT_DER);
+            // When failure to import, leave the loop
+            if ( rv != GNUTLS_E_SUCCESS ) {
+                if (ch_size < 1) {
+                    ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r,
+                            "GnuTLS: Failed to Verify Peer: "
+                            "Failed to import peer certificates.");
+                    ret = HTTP_FORBIDDEN;
+                    goto exit;
+                }
+                ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r,
+                        "GnuTLS: Failed to import some peer certificates. Using %d certificates", 
+                        ch_size);
+                rv = GNUTLS_E_SUCCESS;
+                break;
+            }
+        }
+    } else if (gnutls_certificate_type_get( ctxt->session) == GNUTLS_CRT_OPENPGP) {
+        if (cert_list_size > 1) {
+	    ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r,
 		      "GnuTLS: Failed to Verify Peer: "
 		      "Chained Client Certificates are not supported.");
-	return HTTP_FORBIDDEN;
-    }
+            return HTTP_FORBIDDEN;
+        }
 
-    if (gnutls_certificate_type_get( ctxt->session) == GNUTLS_CRT_X509) {
-        gnutls_x509_crt_init(&cert.x509);
-        rv = gnutls_x509_crt_import(cert.x509, &cert_list[0], GNUTLS_X509_FMT_DER);
-    } else if (gnutls_certificate_type_get( ctxt->session) == GNUTLS_CRT_OPENPGP) {
         gnutls_openpgp_crt_init(&cert.pgp);
         rv = gnutls_openpgp_crt_import(cert.pgp, &cert_list[0], GNUTLS_OPENPGP_FMT_RAW);
+
     } else return HTTP_FORBIDDEN;
  
     if (rv < 0) {
@@ -1035,12 +1098,15 @@ static int mgs_cert_verify(request_rec * r, mgs_handle_t * ctxt)
 
     if (gnutls_certificate_type_get( ctxt->session) == GNUTLS_CRT_X509) {
         apr_time_ansi_put(&expiration_time,
-		      gnutls_x509_crt_get_expiration_time(cert.x509));
+		      gnutls_x509_crt_get_expiration_time(cert.x509[0]));
         apr_time_ansi_put(&activation_time,
-		      gnutls_x509_crt_get_activation_time(cert.x509));
+		      gnutls_x509_crt_get_activation_time(cert.x509[0]));
 
-        rv = gnutls_x509_crt_verify(cert.x509, ctxt->sc->ca_list,
-				ctxt->sc->ca_list_size, 0, &status);
+        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
+            "GnuTLS: Verifying list of  %d certificate(s)", ch_size);
+        rv = gnutls_x509_crt_list_verify(cert.x509, ch_size, 
+                ctxt->sc->ca_list, ctxt->sc->ca_list_size,
+                NULL, 0, 0, &status);
     } else {
         apr_time_ansi_put(&expiration_time,
 		      gnutls_openpgp_crt_get_expiration_time(cert.pgp));
@@ -1076,11 +1142,13 @@ static int mgs_cert_verify(request_rec * r, mgs_handle_t * ctxt)
 	expired = 1;
     }
 
-    if (expiration_time < cur_time) {
-	ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r,
+    if (gnutls_certificate_type_get( ctxt->session) != GNUTLS_CRT_OPENPGP || expiration_time != 0) {
+        if (expiration_time < cur_time) {
+	    ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r,
 		      "GnuTLS: Failed to Verify Peer: "
 		      "Peer Certificate is expired.");
-	expired = 1;
+            expired = 1;
+        }
     }
 
     if (status & GNUTLS_CERT_SIGNER_NOT_FOUND) {
@@ -1102,7 +1170,7 @@ static int mgs_cert_verify(request_rec * r, mgs_handle_t * ctxt)
     }
 
     if (gnutls_certificate_type_get( ctxt->session) == GNUTLS_CRT_X509)
-        mgs_add_common_cert_vars(r, cert.x509, 1,
+        mgs_add_common_cert_vars(r, cert.x509[0], 1,
 			     ctxt->sc->export_certificates_enabled);
     else if (gnutls_certificate_type_get( ctxt->session) == GNUTLS_CRT_OPENPGP)
         mgs_add_common_pgpcert_vars(r, cert.pgp, 1,
@@ -1129,9 +1197,12 @@ static int mgs_cert_verify(request_rec * r, mgs_handle_t * ctxt)
     }
 
   exit:
-    if (gnutls_certificate_type_get( ctxt->session) == GNUTLS_CRT_X509)
-        gnutls_x509_crt_deinit(cert.x509);
-    else if (gnutls_certificate_type_get( ctxt->session) == GNUTLS_CRT_OPENPGP)
+    if (gnutls_certificate_type_get( ctxt->session) == GNUTLS_CRT_X509) {
+        int i;
+        for (i=0; i<ch_size; i++) {
+            gnutls_x509_crt_deinit(cert.x509[i]);
+        }
+    } else if (gnutls_certificate_type_get( ctxt->session) == GNUTLS_CRT_OPENPGP)
         gnutls_openpgp_crt_deinit(cert.pgp);
     return ret;
 
